function [RET, REL, RETREL, ndx] = bb_pr(retrievedbbox, groundtruthbbox, mindetectablesize)
% my own function on evaluating precision & recall. most are copied from
% Label me data tools

% This function returns one point in the precision-recall curve.
% Provided a set of detected bounding boxes, it indicates how may objects
% have been detected (RETREL).
%
% The standard measures for IR are recall and precision. Assuming that:
%
%    * RET is the set of all items the system has retrieved for a specific inquiry;
%    * REL is the set of relevant items for a specific inquiry;
%    * RETREL is the set of the retrieved relevant items 
%
% then precision and recall measures are obtained as follows:
%
%    precision = RETREL / RET
%    recall = RETREL / REL 

% get rid of very small bounding box
gtboxLarge = bb_getlarge(groundtruthbbox, mindetectablesize);

% columns are the number of bounding box.
nGroundTruth = size(groundtruthbbox, 1);
nDetected = size(retrievedbbox, 1);

% Extract the bounding boxes for each target present in the image

% If we detect an object that we thought was too small, we should not
% penalize performance for this. So, we will consider that the only
% relevant targets are the ones with a size larger than the minimal
% detectable size.

REL = size(gtboxLarge, 1);
RET = nDetected;

if REL == 0 
    RETREL = 0;
    return;
end

[cxR, cyR, DxR, DyR] = bb_centersize(retrievedbbox);
[cxO, cyO, DxO, DyO] = bb_centersize(groundtruthbbox);

ndx = [];
for i = 1:RET
    % if detected bbox(i) fits to one groundtruth box (say n), then d(n)=1
    d = sqrt(((cxR(i) - cxO)./DxO).^2 + ((cyR(i) - cyO)./DyO).^2)<.5 ...
        & max(DxO/DxR(i), DxR(i)./DxO)<1.5 ...
        & max(DyO/DyR(i), DyR(i)./DyO)<1.5;
    % find n that fits with bbox(i)
    n = find(d);
    if ~isempty(n)
        ndx(i) = n(1);
    else
        ndx(i) = 0;
    end
end
% how many groundtruth objects are being correctly retrieved, i.e., 
RETREL = sum(unique(ndx)>0); % each object can be detected only once

% If we detect an object that we thought was too small, we should not
% penalize performance for this. So, if we detected one object that was not
% considered before within the relevan set, then we will move it into the
% relevant set:
REL = max(REL, RETREL);



