function [finalWx, finalWy] = nnpull_multilevel(img, frame, count)
[m, n] = size(img);
    
% register img and frames
options.basesSide = 2; % use first two bases
options.nBin = 8;
options.nSide = 2; % partition the image into 2x2 grid, 2x2x8 features.
options.power = 2;
options.nTrain = 10000;
options.nIter = 30;
options.linear = false;

mMinimal = 10;
nMinimal = 10;

resizeRatio = 1.5;
%count = floor(max(log(m / mMinimal) / log(resizeRatio), log(n / nMinimal) / log(resizeRatio)));
%count = 3;

% build the bases in each layers.
currW = n;
currH = m;

bs = cell(count, 1);
for iter = 1:count
    % build Fourier bases 
    bs{iter} = buildFourierBases(currH, currW, options.basesSide);
    currW = floor(currW / resizeRatio);
    currH = floor(currH / resizeRatio);
end;

% then compute the wx and wy
% optionsThis
currW = n;
currH = m;

finalWx = cell(count, 1);
finalWy = cell(count, 1);
eps = 1e-8;

for iter = 1:count
    % resize the template and the frame so that the size [currW, currH] is
    % the size [imageSize, imageSize]
    [patchXs, patchYs] = im_partition(img, currW, currH);
    
    fprintf('Layer = %d/%d...#Patch count = (%d, %d)\n', iter, count, length(patchXs), length(patchYs));
    wcount = zeros(m, n);
    wx = zeros(m, n);
    wy = zeros(m, n);
    
    for i = 1:length(patchXs)
        for j = 1:length(patchYs)
            fprintf('### Patch index = (%d/%d, %d/%d)...\n', i, length(patchXs), j, length(patchYs));
            xsel = patchXs(i):patchXs(i) + currW - 1;
            ysel = patchYs(j):patchYs(j) + currH - 1;
            
            if iter == 1
                wx_crop = zeros(currH, currW);
                wy_crop = zeros(currH, currW);                
            else
                wx_crop = finalWx{iter - 1}(ysel, xsel);
                wy_crop = finalWy{iter - 1}(ysel, xsel);
            end;
            
            bbox = [xsel(1), ysel(1), xsel(end), ysel(end)];
            fprintf('++++++++ generating samples....\n');
            [samplePs, sampleImgs] = nnpull_genSamplesRegion(img, bbox, bs{iter}, options.nTrain, 1, options.power, 1, wx_crop, wy_crop);
            % then extract features
            fprintf('++++++++ extracting features....\n');
            sampleHists = zeros(options.nBin * options.nSide * options.nSide, options.nTrain);
            for k = 1:options.nTrain
                sampleHists(:, k) = im_histFromRawImg(sampleImgs(:, :, k), options.nSide, options.nSide, options.nBin);
            end;
            
            % then do the estimation
            options.wx = wx_crop;
            options.wy = wy_crop;
            fprintf('++++++++ pose estimation....\n');
            [rectifieds, estPs] = nnpull_poseEstHistogram(img, bbox, im_crop(frame, bbox), sampleHists, samplePs, bs{iter}, options);
            % alternatively, we can learn a linear regressor and use it. 
            
            % then update and continue the next 
            [addWx, addWy] = warp_recon(bs{iter}, estPs);
            wx(ysel, xsel) = wx(ysel, xsel) + wx_crop + addWx;
            wy(ysel, xsel) = wy(ysel, xsel) + wy_crop + addWy;
            wcount(ysel, xsel) = wcount(ysel, xsel) + 1;
        end;
    end;
    wx = wx ./ (wcount + eps);
    wy = wy ./ (wcount + eps);
    
    finalWx{iter} = wx;
    finalWy{iter} = wy;
    
    currW = floor(currW / resizeRatio);
    currH = floor(currH / resizeRatio);
end;