function [rectifieds, estPs, closestPatches, sampleImgs, samplesPs, stat, psEachStep] = nnpull_poseEst(img, bbox, imgs, sampleImgs, samplesPs, b, options, estPs)
[nP, nTrain] = size(samplesPs);

options = argutil_setdefaults(options, 'K', 10, 'func', @ms_l2, 'sigma', 1, 'nIter', 15, 'temporal', false, ...
    'normalizeImg', false, 'lognormalize', false, 'laplacian', false, 'riseStop', false, 'functype', 'dist', 'isWeighted', true);

if length(size(sampleImgs)) == 4
    [mSmall, nSmall, nChannel, nTrain] = size(sampleImgs);
    [m, n, nChannel, nFrame] = size(imgs);
    
    rectifieds = zeros(m, n, nChannel, nFrame);
    sampleImgsData = reshape(sampleImgs, [mSmall*nSmall*nChannel, nTrain]);
else
    [mSmall, nSmall, nTrain] = size(sampleImgs);
    [m, n, nFrame] = size(imgs);    
    
    rectifieds = zeros(m, n, nFrame);
    sampleImgsData = reshape(sampleImgs, mSmall*nSmall, nTrain);
    nChannel = 1;
end;

if nargout >= 3
    if nChannel > 1
        closestPatches = zeros(m, n, nChannel, nFrame);
    else
        closestPatches = zeros(m, n, nFrame);
    end;
end;


if ~exist('estPs', 'var');
    estPs = zeros(b.nC, nFrame);
end;

errs = zeros(options.nIter, 1);
psEachStep = zeros(b.nC, options.nIter, nFrame);

if nargout == 4
    stat = zeros(options.nIter, nFrame);
end;

imgPatch = im_crop(img, bbox);
meanPatch = mean(imgPatch(:));
stdPatch = std(imgPatch(:));

if options.lognormalize
    sampleImgsData = log(sampleImgsData);
    meanPatch = mean(log(imgPatch(:)));
end;

if options.laplacian
    lap = fspecial('laplacian');
    sampleImgsData = reshape(imfilter(reshape(sampleImgsData, [mSmall, nSmall, nTrain]), lap), [mSmall*nSmall, nTrain]);
end;

for i = 1:nFrame
    fprintf(1, 'Deal with image #%d...', i);
    if nChannel > 1
        origFrame = imgs(:, :, :, i);
    else
        origFrame = imgs(:, :, i);
    end;
    
    if options.normalizeImg
        meanThis = mean(origFrame(:));
        stdThis = std(origFrame(:));
        origFrame = (origFrame - meanThis) / stdThis * stdPatch + meanPatch;
    end;
    if options.lognormalize
        logImg = log(origFrame);
        meanThis = mean(logImg(:));
        origFrame = logImg - meanThis + meanPatch;
    end;
    
    if options.laplacian
        origFrame = imfilter(origFrame, lap);
    end;
    
    if isstruct('estPs');
        for c = 1:nChannel
            [means, rectified(:, :, c), w] = warp_pullback(origFrame(:, :, c), estPs.Xs(:, :, i), estPs.Ys(:, :, i));
        end;
        ps = zeros(b.nC, 1);
    elseif sum(estPs(:, i).^2) ~= 0
        for c = 1:nChannel
            [means, rectified(:, :, c), w] = warp_pullback(origFrame(:, :, c), b, estPs(:, i));
        end;
        ps = estPs(:, i);
    else
        rectified = origFrame;
        w = ones(m, n);
        ps = zeros(b.nC, 1);
    end;
    
    errs(:) = inf;
    
    for k = 1:options.nIter
        % find nn
    %    feature = ext_v1(rectified, [], 1, 8, 8, 1.0);
        if mSmall ~= m || nSmall ~= n
            rectifiedSmall = imresize(rectified, [mSmall, nSmall]);
            w = imresize(w, [mSmall, nSmall]);
        else
            rectifiedSmall = rectified;
        end;
        % nearest neighbors..
        w = repmat(w, [1, 1, nChannel]);
        [currEstP, indices] = ml_knnreg(sampleImgsData(w(:) > 0.5, :), samplesPs, rectifiedSmall(w(:) > 0.5), options);
        if k == 1 && nargout >= 3
            psOne = samplesPs(:, indices(1));
            if nChannel > 1
                for c = 1:nChannel
                    closestPatches(:, :, c, i) = warp_simulateb_region(img(:, :, c), bbox, b, psOne);
                end;
            else
                closestPatches(:, :, i) = warp_simulateb_region(img, bbox, b, psOne);
            end;
        end;
%        [currEstP, indices] = ml_knnreg(sampleImgs, samplesPs, rectified(:), kNN, @(v1, v2)(ms_mi(v1, v2, 2)), sigma);
%         if norm(currEstP) < norm(estPs(:, i)) / 10
%             fprintf(1, 'Linear area...');
%             options.sigmas = 2*0.5.^(0:9);
%             options.ps = estPs(:, i);
%             [estPs(:, i), rectifiedTemp] = distortEst(reshape(sampleImgs(:, 1), m, n), imgs(:, :, i), b, options);
%             rectified = rectifiedTemp{1, end};
%             break;
%         end;
    %    [currEstP, allindices] = ml_knnreg(features, samplesPs, feature, kNN, @ms_l2, sigma);
        ps = ps + currEstP;
        [wx, wy] = warp_recon(b, ps);
        if isstruct('estPs')
            wx = wx + estPs.Xs(:, :, i);
            wy = wy + estPs.Ys(:, :, i);
        end;
        for c = 1:nChannel
            [means, rectified(:, :, c), w] = warp_pullback(origFrame(:, :, c), wx, wy);
        end;
        % verify it. 
        simulated = warp_simulateb_region(img, bbox, b, ps);
        
        if options.lognormalize
            simulated = log(simulated);
        end;
        if options.laplacian
            simulated = imfilter(simulated, lap);
        end;
        
        errs(k) = sum(colize(simulated - origFrame).^2);

        psEachStep(:, k, i) = ps;
        
        if options.riseStop
            if k > 1 && errs(k) > errs(k - 1)
                break;
            end;
        end;
    end;
    % find the best iteration
    [minErr, minIter] = min(errs);
    %minIter = nIter;
    ps = psEachStep(:, minIter, i);
    [wx, wy] = warp_recon(b, ps);
    if isstruct('estPs')
        wx = wx + estPs.Xs(:, :, i);
        wy = wy + estPs.Ys(:, :, i);
    end;
    
    fprintf(1, 'minIter = %d', minIter);
    
    stat(:, i) = errs;
    
    if nChannel > 1
        for c = 1:nChannel
            [means, rectified(:, :, c)] = warp_pullback(imgs(:, :, c, i), wx, wy);
        end;
    else
        [means, rectified(:, :, c)] = warp_pullback(imgs(:, :, i), wx, wy);
    end;
    
%     if recordEachStep
%         rectSteps(:, :, nIter+1, i) = rectified;
%     end;
    if nChannel > 1
        rectifieds(:, :, :, i) = rectified;
    else
        rectifieds(:, :, i) = rectified;
    end;
    if isstruct('estPs')
        estPs.Xs = wx;
        estPs.Ys = wy;
    else
        estPs(:, i) = ps;
    end;
    if options.temporal
        fprintf(1, '     *** new sample added...');
        % put a new sample into this set
        samplesPs = cat(2, samplesPs, ps);
        newSample = imresize(warp_simulateb_region(img, bbox, b, ps), [mSmall, nSmall]);
        
        if nChannel > 1
            sampleImgs = cat(4, sampleImgs, newSample);
        else
            sampleImgs = cat(3, sampleImgs, newSample);
        end;
        
        if options.lognormalize
            newSample = log(newSample);
        end;
        if options.laplacian
            newSample = imfilter(reshape(newSample, [mSmall, nSmall]), lap);
            newSample = newSample(:);
        end;
        
        sampleImgsData = cat(2, sampleImgsData, newSample(:));
        nTrain = nTrain + 1;
    end;
    
    fprintf(1, '\n');
end;
return;