function [rectifieds, estDeforms, rectSteps] = nnpull_poseEstFreeForm(sourceImg, imgs, nSamplePerIter, smoothSigma, nIter)
% free-form estimation
kNN = 10;
sigma = 5;

[m, n, nFrame] = size(imgs);
rectifieds = zeros(m, n, nFrame);
estDeforms.Xs = zeros(m, n, nFrame);
estDeforms.Ys = zeros(m, n, nFrame);

recordEachStep = false;

if nargout >= 3
    rectSteps = zeros(m, n, nIter+1, nFrame);
    recordEachStep = true;
end;

% mStages = round(linspace(m/10, m, nIter));
% nStages = round(linspace(n/10, n, nIter));
mStages = repmat(round(m/10), 1, nIter);
nStages = repmat(round(n/10), 1, nIter);

for i = 1:nFrame
    fprintf(1, 'Deal with image #%d...', i);
    rectified = imgs(:, :, i);
    
    for k = 1:nIter
        if recordEachStep
            rectSteps(:, :, k, i) = rectified;
        end;
        
        rectifiedResized = imresize(rectified, [mStages(k), nStages(k)]);
        resizedSize = mStages(k)*nStages(k);
        
        [distorted, warp] = deform_generator(sourceImg, rectified, nSamplePerIter, mStages(k), nStages(k), smoothSigma);
        distorted = reshape(distorted, [resizedSize, nSamplePerIter]);
        warps = [reshape(warp.Xs, [m*n, nSamplePerIter]); reshape(warp.Ys, [m*n, nSamplePerIter])];
        % find nn
        [currEstWarps, indices] = ml_knnreg(distorted, warps, rectifiedResized(:), kNN, @ms_l2, sigma);
        
        wx = reshape(currEstWarps(1:m*n), m, n);
        wy = reshape(currEstWarps(m*n+1:end), m, n);
        
        estDeforms.Xs(:, :, i) = estDeforms.Xs(:, :, i) + wx;
        estDeforms.Ys(:, :, i) = estDeforms.Ys(:, :, i) + wy;
        [means, rectified] = warp_pullback(imgs(:, :, i), estDeforms.Xs(:, :, i), estDeforms.Ys(:, :, i));
    end;
    if recordEachStep
        rectSteps(:, :, nIter+1, i) = rectified;
    end;
    rectifieds(:, :, i) = rectified;
    fprintf(1, '\n');
end;

% function [distorted, warps] = deform_generator(sourceImg, destImg, nSamples, smoothSigma)
% [m, n] = size(sourceImg);
% diffImg = im_blur(abs(sourceImg - destImg), smoothSigma, 'replicate');
% % find maximum difference part. and try sampling on this region to get rid of curse of dimensionality
% % or we can just do sampling.
% [maxVal, maxIndex] = max(diffImg(:));
% scale = 400.0;
% 
% % scale = 5000.0;
% % 
% % sampleImg = zeros(m, n);
% % sampleImg(maxIndex) = scale;
% % sampleImg = im_blur(sampleImg, smoothSigma, 'replicate');
% % 
% % warps.Xs = reshape(sampleImg(:) * randn(1, nSamples), [m, n, nSamples]);
% % warps.Ys = reshape(sampleImg(:) * randn(1, nSamples), [m, n, nSamples]);
% maskImg = zeros(m, n);
% maskImg(maxIndex) = 1;
% maskImg = im_blur(maskImg, smoothSigma, 'replicate');
% maskImg = maskImg ./ max(maskImg(:)) * scale;
% 
% warps.Xs = im_blur(randn(m, n, nSamples), smoothSigma, 'replicate') .* repmat(maskImg, [1, 1, nSamples]);
% warps.Ys = im_blur(randn(m, n, nSamples), smoothSigma, 'replicate') .* repmat(maskImg, [1, 1, nSamples]);
% 
% distorted = warp_simulate(sourceImg, warps);

function [distortedsmall, warps] = deform_generator(sourceImg, destImg, nSamples, mSmall, nSmall, smoothSigma)
[m, n] = size(sourceImg);

scale = 15.0;

xs = im_blur(randn(mSmall, nSmall, nSamples), smoothSigma) * scale;
ys = im_blur(randn(mSmall, nSmall, nSamples), smoothSigma) * scale;

warps.Xs = zeros(m, n, nSamples);
warps.Ys = zeros(m, n, nSamples);

for i = 1:nSamples
    warps.Xs(:, :, i) = imresize(xs(:, :, i), [m, n], 'bicubic');
    warps.Ys(:, :, i) = imresize(ys(:, :, i), [m, n], 'bicubic');
end;

distorted = warp_simulate(sourceImg, warps);

distortedsmall = zeros(mSmall, nSmall, nSamples);
for i = 1:nSamples
    distortedsmall(:, :, i) = imresize(distorted(:, :, i), [mSmall, nSmall]);
end;