function [rectifieds, estPs, psEachStep] = nnpull_poseEstHistogram(img, bbox, imgs, sampleImgs, sampleHists, samplesPs, b, optionsThis)
[nP, nTrain] = size(samplesPs);

additional_w = false;
if isfield(optionsThis, 'wx') && isfield(optionsThis, 'wy')
    additional_w = true;
end;

optionsThis.linear = isfield(optionsThis, 'linear') && optionsThis.linear;

if optionsThis.linear
    % learn a linear regressor. 
    nDim = size(sampleHists, 1);
    eps = 1e-7;
    X = [ones(nTrain, 1), sampleHists'];
    betas = inv(X'*X + eps * eye(nDim + 1)) * (X' * samplesPs');
    options = [];
else
    if isfield(optionsThis, 'K') && ~isempty(optionsThis.K)
        options.K = optionsThis.K;
    else
        options.K = 10;
    end;
    %options.sigma = 5;
    options.sigma = 1;
    if isfield(optionsThis, 'func')
        options.func = optionsThis.func;
    else
        options.func = @ms_l2;
    end;

    options.functype = 'dist';
    options.isWeighted = true;
end;

[m, n, nFrame] = size(imgs);
rectifieds = zeros(m, n, nFrame);

errs = zeros(optionsThis.nIter, 1);

estPs = zeros(b.nC, nFrame);
psEachStep = zeros(b.nC, optionsThis.nIter, nFrame);

for i = 1:nFrame
    fprintf(1, 'Deal with image #%d...', i);
    rectified = imgs(:, :, i);
    ps = zeros(b.nC, 1);
    
    errs(:) = inf;
    
    for k = 1:optionsThis.nIter
        % find nn
        feature = im_histFromRawImg(rectified, optionsThis.nSide, optionsThis.nSide, optionsThis.nBin);
        % 
        if optionsThis.linear
            currEstP = ([1, feature'] * betas)';
        else
            [currEstP, indices] = ml_knnreg(sampleHists, samplesPs, feature, options);
        end;
        
        ps = ps + currEstP;
        [wx, wy] = warp_recon(b, ps);
        [means, rectified] = warp_pullback(imgs(:, :, i), wx, wy);
        % verify it. 
        if additional_w
            simulated = warp_simulate_region(img, bbox, wx + optionsThis.wx, wy + optionsThis.wy);
        else
            simulated = warp_simulate_region(img, bbox, wx, wy);
        end;
        
        errs(k) = sum(colize(simulated - imgs(:, :, i)).^2);
        psEachStep(:, k, i) = ps;
        
        if isfield(options, 'riseStop') && options.riseStop
            if k > 1 && errs(k) > errs(k - 1)
                break;
            end;
        end;
    end;
    % find the best iteration
    [minErr, minIter] = min(errs);
    % minIter = optionsThis.nIter;
    ps = psEachStep(:, minIter, i);
    [wx, wy] = warp_recon(b, ps);
    fprintf(1, 'minIter = %d', minIter);
    
    [means, rectified] = warp_pullback(imgs(:, :, i), wx, wy);
    
    rectifieds(:, :, i) = rectified;
    estPs(:, i) = ps;
    if isfield(options, 'temporal') && options.temporal
        % put a new sample into this set
        samplesPs = cat(2, samplesPs, ps);
        sampleImgs = cat(3, sampleImgs, imresize(warp_simulateb_region(img, bbox, b, ps), [mSmall, nSmall]));
        nTrain = nTrain + 1;
    end;
    
    fprintf(1, '\n');
end;