function [sampleImgs, samplesPs] = nnpull_resample(sourceImg, destImg, b, nSample)
diffImg = abs(destImg - sourceImg);
% some blurring
sigma = 4.0; % 4.0
diffImg = im_blur(diffImg, sigma);
% rescale it to between [0, 1]
eps = 1e-2;
diffImg = diffImg ./ max(diffImg(:)) + eps;
diffImg = diffImg / 4;

% probabilistic model. 
%   prior      = p^T * \Sigma_0^{-1} * p (independent variance on p space)
%   likelihood = (Wp)^T * \Sigma_L^{-1} * Wp (independent variance on Wp space, or image space)
%   posterior  = p^T * (\Sigma_0^{-1} + W^T\Sigma_L^{-1}W) * p
% and then sample the posterior. 

sigmaSqr0 = mean(b.variances);

W = [b.bx; b.by];
sigmaLInv = 1 ./ ((diffImg + eps).^2) / sigmaSqr0;

% b.variance is the variance in each dimension.
posteriorPrecision = W' * (repmat(sigmaLInv(:), 2, b.nC) .* W) + diag(1 ./ b.variances);
%posteriorPrecision = diag(1 ./ b.variances);
posteriorSigma = inv(posteriorPrecision);
% posterior = p^TR^T * Rp
R = chol(posteriorSigma);

% then do sampling
samplesPs = R * randn(b.nC, nSample);
sampleImgs = warp_simulateb(sourceImg, b, samplesPs);