function points = nnpull_tracking(frames, initpoints, b, options)
options = argutil_setdefaults(options, 'refinement', false, 'nSample', 300);
% start with the first frame, 
nPoint = size(initpoints, 1);
nFrame = size(frames, 3);

template = frames(:, :, 1);
% for each patch, use the bases to generate some images
optionsCrop.check = true;
optionsCrop.tolerate = false;
optionsCrop.expand = false;

fprintf(1, 'Sample size = %d\n', options.nSample);
% then we can do the tracking
points = zeros(nPoint, 2, nFrame);

points(:, :, 1) = initpoints(:, :, 1);

for i = 1:nPoint
    bboxAll = bb_fromCenterSize(initpoints(i, :, 1), [b.h, b.w]);
    fprintf(1, ['bbox for track point #' num2str(i) ' : ' mat2str(bboxAll) '\n']);
    [samplePsAll, sampleImgsAll] = ...
        nnpull_genSamplesRegion(template, bboxAll, b, options.nSample, 1, 2, 1);
    
    curr_point = initpoints(i, :, 1);
    
    for k = 2:nFrame
        fprintf(1, 'tracking frame #%d\n', k);
        % extract the patch using the previous estimation of the current point
        if options.refinement
            p = round(initpoints(i, :, k));
        else
            p = curr_point;
        end;
        
        bbox = bb_fromCenterSize(p, [b.h, b.w]);
        
        patch = im_crop(frames(:, :, k), bbox, optionsCrop);
        
        if isempty(patch)
            curr_point(:) = 0;
            continue;
        end;
              
        [rectifieds, estPs, closestPatches, sampleImgsAll, samplePsAll] = ...
            nnpull_poseEst(template, bboxAll, patch, sampleImgsAll, samplePsAll, b, options);
        % use estPs to find how much we should proceed given the current positions..
        [wx, wy] = warp_recon(b, estPs);
        [iwx, iwy] = warp_inverse(wx, wy);
        
        curr_point(1) = round(p(1) + iwx(floor(b.h / 2), floor(b.w / 2)));
        curr_point(2) = round(p(2) + iwy(floor(b.h / 2), floor(b.w / 2)));
        
        points(i, :, k) = curr_point;
    end;
end;
