function [meanImg, reconEach, w] = warp_pullback(imgs, b, ps, sigma)
if ~exist('sigma', 'var')
    sigma = 1;
end;
[m, n, nFrames] = size(imgs);
if isstruct(b)
    nFrames2 = size(ps, 2);
else
    nFrames2 = size(b, 3);
end;
isSingleImg = false;
if nFrames ~= nFrames2
    % then nFrames == 1
    nFrames = nFrames2;
    isSingleImg = true;
end;

[X, Y] = meshgrid(1:n, 1:m);
X = reshape(X, m*n, 1);
Y = reshape(Y, m*n, 1);

if nargout >= 2
    reconEach = zeros(m*n, nFrames);
%   warning !! dangerous changes.
%   reconEach = ones(m*n, nFrames) * 0.5;
    w = zeros(m*n, nFrames);
end;

meanImg = zeros(m*n, 1);
count = zeros(m*n, 1);

nC = size(ps, 1);
for i = 1:nFrames
    if isstruct(b)
        newx = b.bx(:, 1:nC) * ps(:, i) + X;
        newy = b.by(:, 1:nC) * ps(:, i) + Y;
    else
        newx = reshape(b(:, :, i), m*n, 1) + X;
        newy = reshape(ps(:, :, i), m*n, 1) + Y;
    end;
 
    % for mean, no need to do interpolation. simplest method works
    newxInt = round(newx);
    newyInt = round(newy);
    
    valid = (newxInt >= 1 & newxInt <= n & newyInt >= 1 & newyInt <= m);
    % note indices are repeatable.
    indices = sub2ind([m, n], newyInt(valid), newxInt(valid));
    if isSingleImg
        thisFrame = reshape(imgs, m, n);
    else
        thisFrame = reshape(imgs(:, :, i), m, n);
    end;
    meanImg(indices) = meanImg(indices) + thisFrame(valid);
    count(indices) = count(indices) + 1;
    
    if nargout >= 2
       reconEach(indices, i) = thisFrame(valid);
       w(indices, i) = 1;
%        if mod(i, 10) == 0
%        fprintf(1, 'recon image = %d...\n', i);
%        end;
%        reconEach(:, i) = myinterp(newx, newy, thisFrame(:), Xdst, Ydst);
    end;
end;

eps = 1e-15;
meanImg = reshape(meanImg ./ (count + eps), m, n);
if nargout >= 2
    reconEach = reshape(reconEach, [m, n, nFrames]);
    w = reshape(w, [m, n, nFrames]);
    % then do the weighted smoothing. 
    gf = fspecial('gaussian', ceil([3*sigma, 3*sigma]), sigma);
%    gf = fspecial('gaussian', [5, 5], 1.5);
%    gf = fspecial('gaussian', [11, 11], 2.5);

    for i = 1:nFrames
        reconEachBlurred = imfilter(reconEach(:, :, i) .* w(:, :, i), gf, 'same');
        wBlurred = imfilter(w(:, :, i), gf, 'same');
        reconEachBlurred = reconEachBlurred ./ (wBlurred + eps);
        
        reconEachFinal = reconEach(:, :, i);
        reconEachFinal(w(:, :, i) == 0) = reconEachBlurred(w(:, :, i) == 0);
        
        reconEach(:, :, i) = reconEachFinal;
    %   reconEach(w <= 1) = reconEach(w <= 1) .* w(w<=1) + reconEachBlurred(w <= 1) .* (1 - w(w<=1));
    %   reconEach = reconEachBlurred;
    end;
end;