function finalBlurred = warp_pullbackOn(canvasSize, patch, b, ps, sx, sy, sigma)
if ~exist('sigma', 'var')
    sigma = 1;
end;
m = canvasSize(1);
n = canvasSize(2);

[mPatch, nPatch] = size(patch);
if ~isstruct(b)
    nFrames = size(ps, 3);
else
    [nC, nFrames] = size(ps);
end;

[X, Y] = meshgrid(1:nPatch, 1:mPatch);
X = reshape(X, mPatch*nPatch, 1);
Y = reshape(Y, mPatch*nPatch, 1);

finalImg = zeros(m, n);
count = zeros(m, n);

for i = 1:nFrames
    if ~isstruct(b)
        newx = reshape(b(:, :, i), nPatch*mPatch, 1) + X + sx(i) - 1;
        newy = reshape(ps(:, :, i), nPatch*mPatch, 1) + Y + sy(i) - 1;
    else
        newx = b.bx(:, 1:nC) * ps(:, i) + X + sx(i) - 1;
        newy = b.by(:, 1:nC) * ps(:, i) + Y + sy(i) - 1;
    end;
 
    % for mean, no need to do interpolation. simplest method works
    newxInt = round(newx);
    newyInt = round(newy);
    
    valid = (newxInt >= 1 & newxInt <= n & newyInt >= 1 & newyInt <= m);
    % note indices are repeatable.
    indices = sub2ind([m, n], newyInt(valid), newxInt(valid));
    
    finalImg(indices) = finalImg(indices) + patch(valid);
    count(indices) = count(indices) + 1;
end;

% then do the weighted smoothing. 
gf = fspecial('gaussian', ceil([3*sigma, 3*sigma]), sigma);
%    gf = fspecial('gaussian', [5, 5], 1.5);
%    gf = fspecial('gaussian', [11, 11], 2.5);

eps = 1e-15;
finalBlurred = imfilter(finalImg, gf, 'same');
countBlurred = imfilter(count, gf, 'same');
finalBlurred = finalBlurred ./ (countBlurred + eps);